"""Quantum Inspire library

Copyright 2019 QuTech Delft

qilib is available under the [MIT open-source license](https://opensource.org/licenses/MIT):

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation the
rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit
persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
from typing import Any, Dict, List, Optional

from qilib.utils.storage.interface import (NoDataAtKeyError,
                                           NodeAlreadyExistsError,
                                           StorageInterface)


class StorageMemory(StorageInterface):
    """ Reference implementation of StorageInterface with in-memory backend.

    Implements a storage tree as an in-memory dictionary.
    """

    class __Leaf:
        def __init__(self, data: Any) -> None:
            self.data: Any = data

    class __Node(Dict[str, Any]):
        pass

    def __init__(self, name: str) -> None:
        """ In memory implementation of storage class.

        See also: `StorageInterface`

        """
        super().__init__(name)
        self._data: Dict[str, Any] = {}

    @staticmethod
    def _retrieve_value_from_dict_by_tag(dictionary: Dict[str, Any],
                                         tag: List[str]) -> Any:
        if len(tag) == 0:
            if not isinstance(dictionary, StorageMemory.__Leaf):
                raise NoDataAtKeyError()
            return dictionary.data
        tag_prefix: str = tag[0]
        if tag_prefix not in dictionary:
            raise NoDataAtKeyError(tag)
        return StorageMemory._retrieve_value_from_dict_by_tag(dictionary[tag_prefix], tag[1:])

    @staticmethod
    def _retrieve_nodes_from_dict_by_tag(dictionary: Dict[str, Any],
                                         tag: List[str]) -> List[str]:
        if len(tag) == 0:
            return list(dictionary.keys())
        tag_prefix = tag[0]
        if tag_prefix not in dictionary:
            raise NoDataAtKeyError(tag)
        if isinstance(dictionary[tag_prefix], StorageMemory.__Leaf):
            raise NoDataAtKeyError(tag)

        return StorageMemory._retrieve_nodes_from_dict_by_tag(dictionary[tag_prefix], tag[1:])

    @staticmethod
    def _store_value_to_dict_by_tag(dictionary: Dict[str, Any], tag: List[str], value: Any) -> None:
        if len(tag) == 1:
            if tag[0] in dictionary:
                if isinstance(dictionary[tag[0]], StorageMemory.__Node):
                    raise NodeAlreadyExistsError()
            dictionary[tag[0]] = StorageMemory.__Leaf(value)
        else:
            if not tag[0] in dictionary:
                dictionary[tag[0]] = StorageMemory.__Node()
            elif isinstance(dictionary[tag[0]], StorageMemory.__Leaf):
                raise NodeAlreadyExistsError(f'Cannot store or replace data, because \'{tag[0]}\' is already a node')

            StorageMemory._store_value_to_dict_by_tag(dictionary[tag[0]], tag[1:], value)

    def load_data(self, tag: List[str]) -> Any:
        if not isinstance(tag, list):
            raise TypeError()
        return self._unserialize(StorageMemory._retrieve_value_from_dict_by_tag(self._data, tag))

    def save_data(self, data: Any, tag: List[str]) -> None:
        StorageMemory._store_value_to_dict_by_tag(self._data, tag, self._serialize(data))

    def get_latest_subtag(self, tag: List[str]) -> Optional[List[str]]:
        child_tags: List[str] = self.list_data_subtags(tag)
        if len(child_tags) == 0:
            return None
        child_tags = sorted(child_tags)
        return tag + [child_tags[-1]]

    def list_data_subtags(self, tag: List[str]) -> List[str]:
        try:
            tags: List[str] = self._retrieve_nodes_from_dict_by_tag(self._data, tag)
        except NoDataAtKeyError:
            return []
        return list(tags)

    def search(self, query: str) -> Any:
        raise NotImplementedError()

    def tag_in_storage(self, tag: List[str]) -> bool:
        tmp_data = self._data
        for partial_tag in tag:
            if partial_tag in tmp_data:
                tmp_data = tmp_data[partial_tag]
            else:
                return False
        return True




"""Quantum Inspire library

Copyright 2019 QuTech Delft

qilib is available under the [MIT open-source license](https://opensource.org/licenses/MIT):

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation the
rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit
persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
from types import ModuleType
from typing import Dict, Tuple, cast, Optional

import qilib
from qilib.configuration_helper import InstrumentAdapter


class InstrumentAdapterFactory:
    """ Creates InstrumentAdapters with corresponding QCoDeS driver.

    Factory class responsible for creating instances of the InstrumentAdapter and the corresponding
    QCoDeS instrument. It is called with the name of the desired InstrumentAdapter sub-class for the
    instrument as well as a dictionary that provides key-value pairs of address specifications.
    """

    instrument_adapters: Dict[Tuple[str, str], InstrumentAdapter] = {}
    _external_adapters: Dict[str, ModuleType] = {}

    @staticmethod
    def add_instrument_adapters(package: ModuleType):
        """ Adds InstrumentAdapters (from an external package)

        Args:
            package: The package that contains InstrumentAdapters
        """

        InstrumentAdapterFactory._external_adapters.update(vars(package))

    @classmethod
    def get_instrument_adapter(cls, instrument_adapter_class_name: str, address: str,
                               instrument_name: Optional[str] = None) -> InstrumentAdapter:
        """ Factory method for creating an InstrumentAdapter with QCoDeS instrument.

        Args:
            instrument_adapter_class_name: Name of the InstrumentAdapter subclass.
            address: Address of the physical instrument.
            instrument_name: An optional name for the underlying instrument.

        Returns:
             An instance of the requested InstrumentAdapter.

        Raises:
            ValueError: If adapter is not found or adapter exist but with a different instrument name.

        """
        instrument_adapter_key = instrument_adapter_class_name, str(address)
        if instrument_adapter_key in cls.instrument_adapters:
            adapter = cls.instrument_adapters[instrument_adapter_key]
            if instrument_name is not None and instrument_name != adapter.instrument.name:
                raise ValueError(f'An adapter exist with different instrument name'
                                 f' \'{adapter.instrument.name}\' != \'{instrument_name}\'')
            return adapter

        adapter = vars(qilib.configuration_helper.adapters).get(instrument_adapter_class_name,
                                                                InstrumentAdapterFactory._external_adapters.get(
                                                                    instrument_adapter_class_name))
        if adapter is None:
            raise ValueError(f"No such InstrumentAdapter {instrument_adapter_class_name}")
        else:
            args = (address, instrument_name) if instrument_name is not None else (address,)
            adapter = cast(InstrumentAdapter, adapter(*args))
            cls.instrument_adapters[instrument_adapter_key] = adapter
            return adapter



"""Quantum Inspire library

Copyright 2019 QuTech Delft

qilib is available under the [MIT open-source license](https://opensource.org/licenses/MIT):

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation the
rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit
persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
from abc import ABC, abstractmethod

from qilib.data_set import DataSet


class DataSetIOReader(ABC):
    """ Abstract base class for data set io readers."""

    DATA = 'data'
    METADATA = 'metadata'
    DATA_ARRAY = 'data_array'
    DATA_ARRAYS = 'data_arrays'
    ARRAY_DATA = 'array_data'
    ARRAY_UPDATES = 'array_updates'

    def __init__(self) -> None:
        """ This is an abstract base class and should not be instantiated directly."""
        self._data_set: DataSet
        self._bound = False

    def bind_data_set(self, data_set: DataSet) -> None:
        """ Binds the DataSet to the DataSetIOReader.

        Args:
            data_set: The object that encompasses DataArrays.

        """
        self._data_set = data_set
        self._bound = True

    @abstractmethod
    def sync_from_storage(self, timeout: float) -> None:
        """ Reads changes from the underlying storage backend and applies them to the bound DataSet.

        Args:
            timeout: Stop syncing after certain amount of time.
        """

    @staticmethod
    @abstractmethod
    def load() -> DataSet:
        """ Opens an existing DataSet from the underlying storage."""



"""Quantum Inspire library

Copyright 2019 QuTech Delft

qilib is available under the [MIT open-source license](https://opensource.org/licenses/MIT):

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation the
rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit
persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
from typing import List, Any, Optional

from qilib.configuration_helper import InstrumentConfigurationSet
from qilib.utils.storage.interface import StorageInterface


class ConfigurationHelper:
    """  Provides access to the instrument configuration database and maintains two InstrumentConfigurationSets."""

    def __init__(self, storage: StorageInterface,
                 active_configuration: Optional[InstrumentConfigurationSet] = None,
                 inactive_configuration: Optional[InstrumentConfigurationSet] = None) -> None:
        """ Instantiate a new ConfigurationHelper.

        Args:
            storage: A StorageInterface that stores configurations.
            active_configuration: An optional instrument configuration set.
            inactive_configuration: An optional inactive instrument configuration set.
        """
        if active_configuration is not None:
            self._active_configuration = active_configuration
        else:
            self._active_configuration = InstrumentConfigurationSet(storage)
        if inactive_configuration is not None:
            self._inactive_configuration = inactive_configuration
        else:
            self._inactive_configuration = InstrumentConfigurationSet(storage)
        self._storage = storage

    def snapshot(self, tag: Optional[List[str]] = None) -> None:
        """ Refresh the active configuration set so it reflects thee actual settings of the devices.

        Args:
            tag: Unique identifier for an InstrumentConfigurationSet. If no tag is provided it will be generated based
                on current time and date.

        """
        self._active_configuration.snapshot(tag)

    @property
    def active_configuration(self) -> InstrumentConfigurationSet:
        """ The configuration set that is currently active."""
        return self._active_configuration

    @property
    def inactive_configuration(self) -> InstrumentConfigurationSet:
        """ The configuration set that is to be applied."""
        return self._inactive_configuration

    def retrieve_inactive_configuration_from_storage(self, tag: List[str]) -> None:
        """ Load configuration set from storage and set as inactive.

        Args:
            tag: Unique identifier for an InstrumentConfigurationSet.

        """
        self._inactive_configuration = InstrumentConfigurationSet.load(tag, self._storage)

    def write_active_configuration_to_storage(self) -> None:
        """ Write the active configuration set to storage."""
        self._active_configuration.store()

    def write_inactive_configuration_to_storage(self) -> None:
        """ Write the inactive configuration set to storage."""
        self._inactive_configuration.store()

    def apply_inactive(self) -> None:
        """ Apply inactive configuration and set as active."""
        self._inactive_configuration.apply()
        self._active_configuration = self._inactive_configuration

    def apply_inactive_delta(self) -> None:
        """ Call apply_delta on inactive configuration set and set it as active."""
        self._inactive_configuration.apply_delta()
        self._active_configuration = self._inactive_configuration

    def apply_inactive_delta_lazy(self) -> None:
        """ Call apply_delta_lazy on inactive configuration set and set it as active."""
        self._inactive_configuration.apply_delta_lazy()
        self._active_configuration = self._inactive_configuration

    def get_tag_by_label(self, label: str) -> Any:
        """ Retrieve tag from storage that has been assigned the given label.

        Args:
            label: A known configuration, e.g., 'online'.

        """
        return self._storage.load_data(['labels', label])

    def label_tag(self, label: str, tag: List[str]) -> None:
        """ Label a tag in storage with a given label.

        Args:
            label: A known configuration, e.g., 'online'.
            tag: The tag to label.

        """
        self._storage.save_data(tag, ['labels', label])



from qilib.configuration_helper.adapters.common_instrument_adapter import CommonInstrumentAdapter
from qilib.configuration_helper.adapters.common_config_instrument_adapter import CommonConfigInstrumentAdapter
from qilib.configuration_helper.adapters.ami430_instrument_adapter import AMI430InstrumentAdapter
from qilib.configuration_helper.adapters.keithley_dmm6500_adapter import Keithley6500InstrumentAdapter
from qilib.configuration_helper.adapters.spi_rack_instrument_adapter import SpiRackInstrumentAdapter
from qilib.configuration_helper.adapters.spi_module_instrument_adapter import SpiModuleInstrumentAdapter
from qilib.configuration_helper.adapters.d5a_instrument_adapter import D5aInstrumentAdapter
from qilib.configuration_helper.adapters.f1d_instrument_adapter import F1dInstrumentAdapter
from qilib.configuration_helper.adapters.hdawg8_instrument_adapter import ZIHDAWG8InstrumentAdapter
from qilib.configuration_helper.adapters.m2j_instrument_adapter import M2jInstrumentAdapter
from qilib.configuration_helper.adapters.s5i_instrument_adapter import S5iInstrumentAdapter
from qilib.configuration_helper.adapters.uhfli_instrument_adapter import ZIUHFLIInstrumentAdapter
from qilib.configuration_helper.adapters.keysight_e8267d_instrument_adapter import KeysightE8267DInstrumentAdapter



"""Quantum Inspire library

Copyright 2019 QuTech Delft

qilib is available under the [MIT open-source license](https://opensource.org/licenses/MIT):

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation the
rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit
persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
from copy import deepcopy
from typing import Any, Union, Dict, Tuple

from qilib.data_set import DataArray
from qilib.data_set.data_set_io_writer import DataSetIOWriter
from qilib.utils.memory_storage_queue import MemoryStorageQueue


class MemoryDataSetIOWriter(DataSetIOWriter):
    """ Allow a DataSet to write changes to an in-memory data storage queue."""

    def __init__(self, storage_queue: MemoryStorageQueue) -> None:
        """ Construct a new instance of MemoryDataSetIOWriter.
            This should not be called directly but a Reader/Writer pair should be created with
            the MemoryDataSetIOFactory.

        Args:
            storage_queue: Fifo shared with a MemoryDataSetIOWriter.
        """
        super().__init__()
        self._storage_queue = storage_queue

    def sync_metadata_to_storage(self, field_name: str, value: Any) -> None:
        self._storage_queue.add_meta_data(field_name, value)

    def sync_data_to_storage(self, index_or_slice: Union[int, Tuple[int]], data: Dict[str, Any]) -> None:
        self._storage_queue.add_data(index_or_slice, data)

    def sync_add_data_array_to_storage(self, data_array: DataArray) -> None:
        self._storage_queue.add_array(deepcopy(data_array))

    def finalize(self) -> None:
        raise NotImplementedError('The finalize function cannot be used with the MemoryDataSetIOWriter!')



from qilib.data_set.data_array import DataArray
from qilib.data_set.data_set import DataSet
from qilib.data_set.memory_data_set_io_factory import MemoryDataSetIOFactory
from qilib.data_set.mongo_data_set_io import MongoDataSetIO
from qilib.data_set.mongo_data_set_io_writer import MongoDataSetIOWriter
from qilib.data_set.mongo_data_set_io_reader import MongoDataSetIOReader



"""Quantum Inspire library

Copyright 2019 QuTech Delft

qilib is available under the [MIT open-source license](https://opensource.org/licenses/MIT):

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation the
rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit
persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
from typing import Dict

from qilib.configuration_helper import InstrumentAdapterFactory, InstrumentAdapter


class SerialPortResolver:
    """ Creates InstrumentAdapters with a serialport connection.

    Factory class responsible for creating instances of the InstrumentAdapter with a serialport connection.
    The serial port number is resolved from the given identifier and used as address to create the
    InstrumentAdapter. The identifier, serial port address combinations are stored in the
    serial_port_identifiers dictionary.
    """

    serial_port_identifiers: Dict[str, str] = {'spirack3': 'COM3',
                                               'spirack4': 'COM1',
                                               'spirack7': 'COM6'}

    @classmethod
    def get_serial_port_adapter(cls, serial_port_adapter_class_name: str, identifier: str) -> InstrumentAdapter:
        """ Factory method for creating an InstrumentAdapter with QCoDeS instrument.

        Gets the instrument adapter given corresponding to the identifier that uniquely
        defines the connection to the QCodes instrument.

        Args:
            serial_port_adapter_class_name: The name of the InstrumentAdapter subclass.
            identifier: The address of the physical instrument.

        Returns:
             An instance of the requested InstrumentAdapter.
        """
        if identifier not in cls.serial_port_identifiers.keys():
            raise ValueError("No such identifier {0} in SerialPortResolver".format(identifier))

        serial_port = cls.serial_port_identifiers[identifier]
        return InstrumentAdapterFactory.get_instrument_adapter(serial_port_adapter_class_name, serial_port)



"""Quantum Inspire library

Copyright 2019 QuTech Delft

qilib is available under the [MIT open-source license](https://opensource.org/licenses/MIT):

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation the
rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit
persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
import base64
from functools import partial
from json import JSONDecoder, JSONEncoder
from typing import Any, Callable, ClassVar, Dict

import numpy as np

from qilib.data_set import MongoDataSetIO

# A callable type for transforming a given argument with a type to another type
TransformFunction = Callable[[Any], Any]


class JsonSerializeKey:
    """ The custom value types for the JSON serializer """
    OBJECT = '__object__'
    CONTENT = '__content__'


class Encoder(JSONEncoder):
    """ A JSON encoder """

    def __init__(self, **kwargs):
        """ Constructs a JSON Encoder """
        super().__init__(**kwargs)

        # creates a new transform table
        self.encoders: ClassVar[Dict[type, TransformFunction]] = {}

    def default(self, o):
        if type(o) in self.encoders:
            return self.encoders[type(o)](o)

        return JSONEncoder.default(self, o)


class Decoder(JSONDecoder):
    """ A JSON decoder """

    def __init__(self):
        """ Constructs a JSON Decoder """
        super().__init__(object_hook=self._object_hook)

        # creates a new transform table
        self.decoders: ClassVar[Dict[str, TransformFunction]] = {}

    def _object_hook(self, obj):
        if isinstance(obj, dict):
            if JsonSerializeKey.OBJECT in obj:
                if obj[JsonSerializeKey.OBJECT] in self.decoders:
                    return self.decoders[obj[JsonSerializeKey.OBJECT]](obj)
                else:
                    raise ValueError()

        return obj


def _encode_bytes(data: Any) -> Dict[str, Any]:
    return {JsonSerializeKey.OBJECT: bytes.__name__, JsonSerializeKey.CONTENT: data.decode('utf-8')}


def _decode_bytes(data: Dict[str, Any]) -> bytes:
    return data[JsonSerializeKey.CONTENT].encode('utf-8')


def _encode_tuple(item):
    return {
        JsonSerializeKey.OBJECT: tuple.__name__,
        JsonSerializeKey.CONTENT: [serializer.encode_data(value) for value in item]
    }


def _decode_tuple(data: Dict[str, Any]) -> tuple:
    return tuple(data[JsonSerializeKey.CONTENT])


def _encode_numpy_number(item):
    return {
        JsonSerializeKey.OBJECT: '__npnumber__',
        JsonSerializeKey.CONTENT: {
            '__npnumber__': base64.b64encode(item.tobytes()).decode('ascii'),
            '__data_type__': item.dtype.str,
        }
    }


def _decode_numpy_number(item):
    obj = item[JsonSerializeKey.CONTENT]
    return np.frombuffer(base64.b64decode(obj['__npnumber__']), dtype=np.dtype(obj['__data_type__']))[0]


def _encode_dataclass(object_: Any, class_name: str) -> Dict[str, Any]:
    """ Encodes a JSON dataclass object

    Args:
        object_: Object to be encoded
        class_name: Object dataclass name

    Returns:
        Dictionary that can be serialized
    """
    return {JsonSerializeKey.OBJECT: class_name,
            JsonSerializeKey.CONTENT: object_.to_dict()}


def _decode_dataclass(data: Any, class_type: type) -> Any:
    """ Decodes a JSON dataclass object

    Args:
        data: Data to be decoded
        class_type: The dataclass type to decode

    Returns
        Object of specified class_type type with decoded data
    """
    return class_type.from_dict(data[JsonSerializeKey.CONTENT])


class Serializer:
    """ A general serializer to serialize data to JSON and vice versa. It allows
     extending the types with a custom encoder and decoder."""

    def __init__(self, encoders: Dict[type, TransformFunction] = None,
                 decoders: Dict[str, TransformFunction] = None):
        """ Creates a serializer

        Args:
            encoders: The default encoders if any
            decoders: The default decoders if any
        """

        self.encoder = Encoder()
        self.decoder = Decoder()

        if encoders is None:
            encoders = {}
        self.encoder.encoders = encoders
        if decoders is None:
            decoders = {}
        self.decoder.decoders = decoders

        self.register(bytes, _encode_bytes, bytes.__name__, _decode_bytes)
        self.register(np.ndarray, MongoDataSetIO.encode_numpy_array, np.array.__name__,
                      MongoDataSetIO.decode_numpy_array)
        self.register(tuple, _encode_tuple, tuple.__name__, _decode_tuple)
        for numpy_integer_type in [np.int16, np.int32, np.int64, np.float16, np.float32, np.float64, np.bool_]:
            self.register(numpy_integer_type, _encode_numpy_number, '__npnumber__', _decode_numpy_number)

    def register(self, type_: type, encode_func: TransformFunction, type_name: str,
                 decode_func: TransformFunction) -> None:
        """ Registers an encoder and decoder for a given type

        An encode function is expected to return a JSON valid type or a dictionary with the keys
        `JsonSerializeKey.OBJECT` (the name of the type) and `JsonSerializeKey.CONTENT` (the JSON valid encoded content)

        Args:
            type_: The type to encode
            encode_func: The transform function for encoding that type
            type_name: The type name to decode
            decode_func: The transform function for decoding
        """

        self.encoder.encoders[type_] = encode_func
        self.decoder.decoders[type_name] = decode_func

    def register_dataclass(self, type_: type) -> None:
        """ Registers an encoder and decoder for a dataclass with a given type

        Args:
            type_: The dataclass type to decode
        """

        type_name = type_.__name__
        encode_function = partial(_encode_dataclass, class_name=type_name)
        decode_function = partial(_decode_dataclass, class_type=type_)
        self.register(type_, encode_function, type_name, decode_function)

    def serialize(self, data: Any) -> str:
        """ Serializes a Python object to JSON

        Args:
            data: Any Python object

        Returns:
            JSON encoded string
        """

        return self.encoder.encode(self.encode_data(data))

    def unserialize(self, data: str) -> Any:
        """ Unserializes a JSON string to a Python object

        Args:
            data: The JSON encoded string

        Returns:
            A Python object decoded from the JSON string
        """

        return self.decoder.decode(data)

    def encode_data(self, data: Any) -> Any:
        """ Recursively transform a Python object and apply transform functions to it

        Args:
            data: Any Python object that can be handled by an encode/transform function for that type

        Returns:
            The transformed data
        """

        if isinstance(data, dict):
            new_dict = {}

            for key, value in data.items():
                new_dict[key] = self.encode_data(value)

            return new_dict

        elif isinstance(data, list):
            new_list = []
            for item in data:
                new_list.append(self.encode_data(self._encode(item)))

            return new_list

        return self._encode(data)

    def _encode(self, data: Any) -> Any:
        type_ = type(data)
        if type_ in self.encoder.encoders:
            return self.encoder.encoders[type_](data)

        return data

    def decode_data(self, data: Any) -> Any:
        """ Recursively transform an object and apply transform functions to it

        Args:
            data: Any Python object that can be handled by an decode/transform function for that type

        Returns:
            The transformed data
        """

        if isinstance(data, dict):
            new_dict = {}
            for key, value in data.items():
                new_dict[key] = self.decode_data(value)

            return self._decode(new_dict)

        if isinstance(data, list):
            new_list = []
            for item in data:
                new_list.append(self.decode_data(item))
            return new_list

        return data

    def _decode(self, data: Any) -> Any:
        if isinstance(data, dict):
            if JsonSerializeKey.OBJECT in data:
                if data[JsonSerializeKey.OBJECT] in self.decoder.decoders:
                    return self.decoder.decoders[data[JsonSerializeKey.OBJECT]](data)
                else:
                    raise ValueError()

        return data


def serialize(data: Any) -> bytes:
    """ Serializes a Python object to JSON using the default serializer. Dictionary keys should be hashable.

    Args:
        data: Any Python object
    Returns:
        JSON encoded string
    """

    return serializer.serialize(data).encode('utf8')


def unserialize(data: bytes) -> Any:
    """ Unserializes a JSON string to a Python object using the default serializer

    Args:
        data: The JSON encoded string
    Returns:
        A Python object decoded from the JSON string
    """

    return serializer.unserialize(data.decode('utf8'))


# The default Serializer to use
serializer = Serializer()



"""Quantum Inspire library

Copyright 2019 QuTech Delft

qilib is available under the [MIT open-source license](https://opensource.org/licenses/MIT):

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation the
rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit
persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
from typing import Optional, Union, Dict, Any, Tuple

from qilib.data_set import DataArray, MongoDataSetIO
from qilib.data_set.data_set_io_reader import DataSetIOReader
from qilib.data_set.data_set_io_writer import DataSetIOWriter


class MongoDataSetIOWriter(DataSetIOWriter):
    """ Allow a DataSet to store changes, and complete DataSet, to a mongodb."""

    def __init__(self, name: Optional[str] = None, document_id: Optional[str] = None,
                 database: str = MongoDataSetIO.DEFAULT_DATABASE_NAME,
                 collection: str = MongoDataSetIO.DEFAULT_COLLECTION_NAME) -> None:
        """ Construct a new instance of MongoDataSetIOWriter. If name is provided, but not found in the database
            a new document is created with that name.

        Args:
            name: DataSet name.
            document_id: _id of the DataSet in the database.
            database: Name of the database.
            collection: Name of the collections.

        Raises:
            DocumentNotFoundError: If document_id is provided but not found in the database.

        """
        super().__init__()
        self._mongo_data_set_io = MongoDataSetIO(name, document_id, database=database, collection=collection)

    def sync_metadata_to_storage(self, field_name: str, value: Any) -> None:
        """ Update or add metadata field to database.

        Args:
            field_name: Field that changed.
            value:  The new value.

        """
        self._is_finalized()
        update_data = {"{}.{}".format(DataSetIOReader.METADATA, field_name): value}
        self._mongo_data_set_io.update_document(update_data)

    def sync_data_to_storage(self, index_or_slice: Union[Tuple[int, ...], int], data: Dict[str, Any]) -> None:
        """ Registers a DataArray update to the database. The change is registered as a change event and is
            applied on finalize().

        Args:
            index_or_slice: The indices of the DataArray to update.
            data: Name of the DataArray to be updated and the new value.

        """
        self._is_finalized()
        update_data = {DataSetIOReader.ARRAY_UPDATES: (index_or_slice, data)}
        self._mongo_data_set_io.append_to_document(update_data)

    def sync_add_data_array_to_storage(self, data_array: DataArray) -> None:
        """ Add or update a DataArray in the database.

        Args:
            data_array: The DataArray to be updated or added.

        """
        self._is_finalized()
        update_data = {"{}.{}".format(DataSetIOReader.DATA_ARRAYS, data_array.name): {
            "name": data_array.name,
            "label": data_array.label,
            "unit": data_array.unit,
            "is_setpoint": data_array.is_setpoint,
            "set_arrays": [array.name for array in data_array.set_arrays],
            "preset_data": MongoDataSetIO.encode_numpy_array(data_array)}}
        self._mongo_data_set_io.update_document(update_data)

    def finalize(self) -> None:
        """ Update the underlying DataSet and close the connection to the database."""
        self._mongo_data_set_io.update_document({DataSetIOReader.ARRAY_UPDATES: []})
        self._mongo_data_set_io.finalize()
        self._finalized = True



"""Quantum Inspire library

Copyright 2019 QuTech Delft

qilib is available under the [MIT open-source license](https://opensource.org/licenses/MIT):

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation the
rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit
persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""

import logging
from abc import ABC, abstractmethod
from datetime import datetime
from typing import Any, Callable, Optional, List, Union


class NoDataAtKeyError(Exception):
    pass


class NodeAlreadyExistsError(Exception):
    pass


class StorageInterface(ABC):
    """ Base class for storage of measurement and calibration results.
        The storage is based on tags (HDF5-like).

        A tag is a list that describes the path in a tree. Data is
        stored at the leafs of the tree.

    Rules:
        nodes cannot contain data
        leaves can be overwritten by leaves
        nodes cannot be overwritten by leaves
    """

    def __init__(self, name: str) -> None:
        """
        Base constructor.

        Args:
            name: Symbolic name for the storage instance.
        """

        self.name: str = name
        self._serialize: Callable[[Any], Any] = lambda x: x
        self._unserialize: Callable[[Any], Any] = lambda x: x
        self.logger: Any = logging.getLogger(self.name)
        self.logger.info('created StorageInterface %s', self.name)

    @staticmethod
    def datetag_part(date_with_time: Optional[datetime] = None) -> str:
        """
        Return string with date.

        Args:
            date_with_time: optional datetime to use instead of `now()`.

        Returns:
             If dt is specified, returns the ISO 8601 formatted date as a string. If
             dt is not specified or None, returns the current date and time formatted
             as a ISO 8601 string.
        """

        if date_with_time is None:
            date_with_time = datetime.now()
        return date_with_time.isoformat()

    @abstractmethod
    def load_data(self, tag: List[str]) -> Any:
        """ Load result from storage.

        Args:
            tag: tag for result to load

        Returns:
            Data found at the node identified by the tag.

        Raises:
            NoDataAtKeyException: if there is no data for the specified tag.
        """

    @staticmethod
    def _tag_to_list(tag: Union[str, List[str]]) -> List[str]:
        """ Convert a str or list tag to list format. """
        if not isinstance(tag, (str, list)):
            raise TypeError('tag should be of type %s' % list)
        if isinstance(tag, str):
            tag = tag.split('/')
        return tag

    @staticmethod
    def _tag_to_string(tag: Union[str, List[str]]) -> str:
        """ Convert a tag to string format. """
        if not isinstance(tag, (str, list)):
            raise TypeError('tag should be of type %s' % list)
        if isinstance(tag, (list,)):
            tag = '/'.join(tag)
        return tag

    @abstractmethod
    def save_data(self, data: Any, tag: List[str]) -> None:
        """ Save data to storage.

        Args:
            data: data to store
            tag: reference tag to store the data
        """
        pass

    @abstractmethod
    def get_latest_subtag(self, tag: List[str]) -> Optional[List[str]]:
        """ Return tag of latest result for a given tag.

        Args:
            tag: reference tag to retrieve latest result

        Returns:
            The tag of the first result for the nodes found at the tag sorted
            in descending order. This implies that if the node at the tag has
            children that are tagged with a datetime string, the tag of the
            most recent result is returned.

            If an invalid tag was specified, None is returned.
        """
        pass

    @abstractmethod
    def list_data_subtags(self, tag: List[str]) -> List[str]:
        """ List available result tags of at tag.

        Args:
            tag: hdf5 tag
        Returns:
            results: List of child tags for tag

        See also:
            load_data

        Notes:
            When an invalid path is specified (for example, a tag
            that addresses a leaf node or a tag that has a component
            that does not exist in the storage hierarchy) an empty
            list is returned.
        """
        pass

    @abstractmethod
    def search(self, query: str) -> Any:
        """ Future implementation of query interface """
        raise NotImplementedError('search interface not yet specified')

    @abstractmethod
    def tag_in_storage(self, tag: List[str]) -> bool:
        """ Check if tag is already in storage
        Args:
            tag: hdf5 tag
        Returns:
            True of tag found in storage, else False.
        """



"""Quantum Inspire library

Copyright 2019 QuTech Delft

qilib is available under the [MIT open-source license](https://opensource.org/licenses/MIT):

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation the
rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit
persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
from typing import Any, Optional

from qcodes.instrument_drivers.QuTech.D5a import D5a

from qilib.configuration_helper.adapters.common_config_instrument_adapter import CommonConfigInstrumentAdapter
from qilib.configuration_helper.adapters.spi_module_instrument_adapter import SpiModuleInstrumentAdapter
from qilib.utils import PythonJsonStructure


class SpanValueError(Exception):
    """ Error when dacs have misconfigured span values."""


DAC_STEP = 10e-3
INTER_DELAY = 0.1
RESET_VOLTAGE = False
MV = True


class D5aInstrumentAdapter(CommonConfigInstrumentAdapter, SpiModuleInstrumentAdapter):

    def __init__(self, address: str, instrument_name: Optional[str] = None) -> None:
        super().__init__(address, instrument_name)
        self._instrument = D5a(self._instrument_name, self._spi_rack, self._module_number, mV=MV,
                               inter_delay=INTER_DELAY,
                               reset_voltages=RESET_VOLTAGE, dac_step=DAC_STEP)
        if self._instrument.span3() != '4v bi':
            raise SpanValueError('D5a instrument has span unequal to "4v bi"')

    def apply(self, config: PythonJsonStructure) -> None:
        """ Applies configuration

        Applies the configuration update to instrument for dac parameters (step, inter_delay and unit)
        And, compares all other configuration value with setter command to the existing configuration
        and raised error in case of mismatch

        Args:
            config: Containing the instrument configuration.

        """
        unit = config['dac1']['unit']
        self._instrument.set_dac_unit(unit)
        dac_parameters = {param: values for param, values in config.items() if param[0:3] == 'dac'}
        for dac, values in dac_parameters.items():
            self._instrument[dac].step = values['step']
            self._instrument[dac].inter_delay = values['inter_delay']
        super().apply(config)

    def _compare_config_values(self, config_value: Any, device_value: Any, parameter: Optional[str] = None) -> bool:
        return bool(config_value != device_value)



"""Quantum Inspire library

Copyright 2019 QuTech Delft

qilib is available under the [MIT open-source license](https://opensource.org/licenses/MIT):

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation the
rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit
persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
from typing import Union, List

from qilib.configuration_helper import InstrumentConfiguration
from qilib.configuration_helper.visitor import Visitor
from qilib.configuration_helper.exceptions import DuplicateTagError
from qilib.utils.storage.interface import StorageInterface


class InstrumentConfigurationSet:
    """ The complete set of configurations for all instruments in the system. """

    STORAGE_BASE_TAG = 'configuration_set'

    def __init__(self, storage: StorageInterface, tag: Union[None, List[str]] = None,
                 instruments: Union[None, List[InstrumentConfiguration]] = None) -> None:
        """ A set of instrument configurations

        Args
            storage: Any storage that implements the StorageInterface
            tag: A unique identifier for a instrument configuration set
            instruments: A list of instrument configurations
        """

        self._storage = storage
        self._tag = [self.STORAGE_BASE_TAG, StorageInterface.datetag_part()] if tag is None else tag
        if instruments is None:
            instruments = []
        self._instruments = instruments


    @property
    def tag(self) -> List[str]:
        """ A unique identifier for this instrument configuration set """

        return self._tag

    @property
    def storage(self) -> StorageInterface:
        """ The storage interface used """

        return self._storage

    @property
    def instruments(self) -> List[InstrumentConfiguration]:
        """ The instrument configurations in this set """

        return self._instruments

    @staticmethod
    def load(tag: List[str], storage: StorageInterface) -> 'InstrumentConfigurationSet':
        """ A factory that creates a new InstrumentConfigurationSet by loading from database

        Args:
            tag: A unique identifier for a instrument configuration set
            storage: Any storage that implements the StorageInterface

        Returns:
            A new InstrumentConfigurationSet loaded from the storage
        """

        # load the document as a list of instruments tags
        tags = storage.load_data(tag)
        instruments = [InstrumentConfiguration.load(instrument_tag, storage) for instrument_tag in tags]

        return InstrumentConfigurationSet(storage, tag, instruments)

    def store(self) -> None:
        """ Saves object to storage

         Raises:
             DuplicateTagError: If this object's tag is already in the database.
        """

        if self._storage.tag_in_storage(self._tag):
            raise DuplicateTagError(f'InstrumentConfiguration with tag \'{self._tag}\' already in storage')

        instruments = []
        for instrument in self.instruments:
            try:
                instrument.store()
            except DuplicateTagError:
                pass
            finally:
                instruments.append(instrument.tag)

        # save the document as a list of instruments tags
        self._storage.save_data(instruments, self._tag)

    def snapshot(self, tag: Union[None, List[str]] = None) -> None:
        """ Updates the configuration set by overwriting the tag and refreshing all underlying instruments """

        self._tag = [self.STORAGE_BASE_TAG, StorageInterface.datetag_part()] if tag is None else tag

        for instrument in self._instruments:
            instrument.refresh()

    def apply(self) -> None:
        """ Uploads the configurations to the instruments """

        for instrument in self.instruments:
            instrument.apply()

    def apply_delta(self) -> None:
        """ Compare configurations with instruments and apply configurations that differs """

        for instrument in self.instruments:
            instrument.apply_delta()

    def apply_delta_lazy(self) -> None:
        """ Compare configurations with instrument drivers last known settings and apply configurations that differs """

        for instrument in self.instruments:
            instrument.apply_delta_lazy()

    def accept(self, visitor: Visitor) -> None:
        """ Accepts a visitor and propagates to instrument configurations.

        Args:
            visitor: An implementation of the Visitor interface.
        
        """
        for instrument in self.instruments:
            instrument.accept(visitor)



"""Quantum Inspire library

Copyright 2019 QuTech Delft

qilib is available under the [MIT open-source license](https://opensource.org/licenses/MIT):

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation the
rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit
persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""

from setuptools import setup


def get_version_number(module: str) -> str:
    """ Extract the version number from the source code.

    Pass the source module that contains the version.py file.
    This version number will be returned as a string.

    Args:
        module: module containing the version.py file

    Returns:
        the version number.
    """
    with open(f'src/{module}/version.py') as f:
        content = f.read()

    return content.split('\'')[1]


def get_long_description() -> str:
    """ Extract the long description from the README file."""
    with open('README.md', encoding='utf-8') as f:
        long_description = f.read()

    return long_description


setup(name='qilib',
      description='Quantum Library for the Quantum Inspire platform',
      long_description=get_long_description(),
      long_description_content_type='text/markdown',
      version=get_version_number('qilib'),
      author='QuantumInspire',
      python_requires='>=3.6',
      package_dir={'': 'src'},
      packages=['qilib', 'qilib.configuration_helper', 'qilib.configuration_helper.adapters',
                'qilib.data_set', 'qilib.utils', 'qilib.utils.storage'],
      classifiers=[
          'Development Status :: 3 - Alpha',
          'Programming Language :: Python :: 3',
          'Programming Language :: Python :: 3.6',
          'Programming Language :: Python :: 3.7'],
      license='Other/Proprietary License',

      install_requires=['spirack>=0.1.8', 'numpy', 'serialize', 'zhinst',
                        'pymongo', 'requests', 'qcodes', 'dataclasses-json'],
      extras_require={
          'dev': ['pytest>=3.3.1', 'coverage>=4.5.1', 'mongomock'],
      })



"""Quantum Inspire library

Copyright 2019 QuTech Delft

qilib is available under the [MIT open-source license](https://opensource.org/licenses/MIT):

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation the
rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit
persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
from typing import Union, Any

from qcodes import Instrument

from qilib.configuration_helper import InstrumentAdapter, InstrumentConfiguration
from qilib.configuration_helper.visitor import Visitor


class InstrumentConfigurationVisitor(Visitor):
    """ Implementation of Visitor interface to visit InstrumentConfiguration."""
    def __init__(self) -> None:
        """ Instantiate a new visitor."""
        self.instruments = []

    def visit(self, element: Union[InstrumentAdapter, InstrumentConfiguration]) -> None:
        """ Visit InstrumentConfiguration and/or InstrumentAdapter.

        Args:
            element: An instance of InstrumentAdapter or InstrumentConfiguration that has been visited.

        """
        if isinstance(element, InstrumentAdapter):
            self._visit_instrument_adapter(element)
        elif isinstance(element, InstrumentConfiguration):
            pass

    def _visit_instrument_adapter(self, instrument_adapter: InstrumentAdapter) -> None:
        self.instruments.append(instrument_adapter.instrument)

    def get_instrument(self, adapter_identifier: str) -> Instrument:
        """ Returns a specific instrument given the adapter name, address combination.

        Args:
            adapter_identifier: The adapter name and address
                                combination (e.g. ZIHDAWG8InstrumentAdapter_DEV8049).

        Raises:
            ValueError: If an adapter identifier is given which is not in the loaded adapters.

        Returns:
            The qcodes instrument which belongs to the adapter identifier.
        """
        instrument = next((i for i in self.instruments if i.name == adapter_identifier), None)
        if instrument is None:
            raise ValueError(f'No adapter with identifier {adapter_identifier} found!')
        return instrument



"""Quantum Inspire library

Copyright 2019 QuTech Delft

qilib is available under the [MIT open-source license](https://opensource.org/licenses/MIT):

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation the
rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit
persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""

from operator import itemgetter
from typing import Any, List, Optional, Union

import numpy as np
from bson import ObjectId
from bson.codec_options import TypeCodec, CodecOptions, TypeRegistry
from pymongo import MongoClient

from qilib.data_set.mongo_data_set_io import MongoDataSetIO, NumpyKeys
from qilib.utils.serialization import Serializer, serializer as _serializer
from qilib.utils.storage.interface import (NoDataAtKeyError,
                                           NodeAlreadyExistsError,
                                           StorageInterface)


class NumpyArrayCodec(TypeCodec):  # type: ignore

    @property
    def python_type(self) -> Any:
        return np.ndarray

    def transform_python(self, value: Any) -> Any:
        return MongoDataSetIO.encode_numpy_array(value)

    @property
    def bson_type(self) -> Any:
        return dict

    def transform_bson(self, value: Any) -> Any:
        if NumpyKeys.OBJECT in value and value[NumpyKeys.OBJECT] == np.array.__name__:
            return MongoDataSetIO.decode_numpy_array(value)

        return value


class StorageMongoDb(StorageInterface):
    """Reference implementation of StorageInterface with an mongodb backend

    Implements a storage tree in a MongoDB collection
    """

    def __init__(self, name: str, host: str = 'localhost', port: int = 27017, database: str = '',
                 serializer: Union[Serializer, None] = None):
        """MongoDB implementation of storage class

        See also: `StorageInterface`

        Args:
            name: Symbolic name for the storage instance.
            host: MongoDB host
            port: MongoDB port
            database: The database to use, if empty the name of the storage is used
        """
        super().__init__(name)

        type_registry = TypeRegistry([NumpyArrayCodec()])
        codec_options = CodecOptions(type_registry=type_registry)
        self._client = MongoClient(host, port)
        self._db = self._client.get_database(database or name, codec_options=codec_options)
        self._collection = self._db.get_collection('storage')

        if serializer is None:
            serializer = _serializer
        self._serialize = serializer.encode_data
        self._unserialize = serializer.decode_data

    def _get_root(self) -> ObjectId:
        """Get or create a root node if it doesn't exist yet

        Returns:
            An ObjectID of the root node
        """

        node = self._collection.find_one({'tag': '', 'parent': {'$exists': False}})

        if node is not None:
            return node['_id']
        else:
            return self._collection.insert_one({'tag': ''}).inserted_id

    def _retrieve_nodes_by_tag(self, tag: List[str], parent: ObjectId) -> List[str]:
        """Traverse the tree and list the children of a given tag

        Args:
            tag: The node tag
            parent: The ObjectID of the node's parent

        Returns:
            A list of names of the children
        """

        if len(tag) == 0:
            return list(map(itemgetter('tag'),
                            self._collection.find({'parent': parent, 'tag': {'$exists': True}}, {'value': 0})))

        else:
            doc = self._collection.find_one({'parent': parent, 'tag': tag[0]})
            if doc is None:
                raise NoDataAtKeyError(f'Tag "{tag[0]}" cannot be found')
            elif 'value' in doc:
                raise NoDataAtKeyError(f'Tag "{tag[0]}" is not a node')
            else:
                return self._retrieve_nodes_by_tag(tag[1:], doc['_id'])

    def _retrieve_value_by_tag(self, tag: List[str], parent: ObjectId) -> Any:
        """Traverse the tree and give the value a given leaf tag

        Args:
            tag: The leaf tag
            parent: The ObjectID of the leaf's parent

        Returns:
            Data held by the leaf
        """

        if len(tag) == 1:
            doc = self._collection.find_one({'parent': parent, 'tag': tag[0]})
            if doc is None:
                raise NoDataAtKeyError(f'Tag "{tag[0]}" cannot be found')
            elif 'value' not in doc:
                raise NoDataAtKeyError(f'Tag "{tag[0]}" is not a leaf')
            else:
                return doc['value']

        else:
            doc = self._collection.find_one({'parent': parent, 'tag': tag[0], 'value': {'$exists': False}})
            if doc is None:
                raise NoDataAtKeyError(f'Tag "{tag[0]}" cannot be found')
            else:
                return self._retrieve_value_by_tag(tag[1:], doc['_id'])

    def _store_value_by_tag(self, tag: List[str], data: Any, parent: ObjectId) -> None:
        """ Store a value at a given tag

        Args:
            tag: The tag
            data: Data to store
            parent: An ObjectID of the node's parent
        """

        if len(tag) == 1:
            doc = self._collection.find_one({'parent': parent, 'tag': tag[0]})
            if doc:
                if 'value' not in doc:
                    raise NodeAlreadyExistsError(f'Tag "{tag[0]}" is not a leaf')
                else:
                    self._collection.update_one({'parent': parent, 'tag': tag[0]},
                                                {'$set': {'value': data}})
            else:
                self._collection.insert_one({'parent': parent, 'tag': tag[0], 'value': data})

        else:
            doc = self._collection.find_one({'parent': parent, 'tag': tag[0]})
            if doc is None:
                parent = self._collection.insert_one({'parent': parent, 'tag': tag[0]}).inserted_id
            else:
                if 'value' in doc:
                    raise NodeAlreadyExistsError(f'Tag "{tag[0]}" is a leaf')
                else:
                    parent = doc['_id']

            self._store_value_by_tag(tag[1:], data, parent)

    def load_data(self, tag: List[str]) -> Any:
        if not isinstance(tag, list):
            raise TypeError('Tag should be a list of strings')

        if len(tag) == 0:
            raise NoDataAtKeyError('Tag cannot be empty')

        return self._unserialize(self._decode_data(self._retrieve_value_by_tag(tag, self._get_root())))

    def save_data(self, data: Any, tag: List[str]) -> None:
        if not isinstance(tag, list):
            raise TypeError('Tag should be a list of strings')

        self._store_value_by_tag(tag, self._encode_data(self._serialize(data)), self._get_root())

    def get_latest_subtag(self, tag: List[str]) -> Optional[List[str]]:
        child_tags = sorted(self.list_data_subtags(tag))
        if len(child_tags) == 0:
            return None

        return tag + [child_tags[-1]]

    def list_data_subtags(self, tag: List[str]) -> List[str]:
        try:
            tags = self._retrieve_nodes_by_tag(tag, self._get_root())
        except NoDataAtKeyError:
            tags = []

        return tags

    def search(self, query: str) -> Any:
        raise NotImplementedError()

    def tag_in_storage(self, tag: List[str]) -> bool:
        parent = self._get_root()
        for tag_part in tag:
            doc = self._collection.find_one({'parent': parent, 'tag': tag_part}, {'value': 0})
            if doc is None:
                return False
            else:
                parent = doc['_id']
        return True

    @staticmethod
    def _encode_int(value: int) -> str:
        """Encodes an int value to a string

        Args:
            value: An integer

        Returns:
            An encoded integer
        """

        return f'_integer[{value}]'

    @staticmethod
    def _is_encoded_int(value: str) -> bool:
        """Checks if the given value is a properly encoded integer

        Args:
            value: The encoded value

        Returns:
            True if an integer can be decoded, False otherwise
        """

        if not value.startswith('_integer[') or not value.endswith(']'):
            return False

        value = value[len('_integer['):-1]
        return value[1:].isdigit() if value.startswith('-') else value.isdigit()

    @staticmethod
    def _decode_int(value: str) -> int:
        """Parses an integer from the encoded value

        Args:
            value: The encoded integer

        Returns:
            The decoded integer
        """

        if not StorageMongoDb._is_encoded_int(value):
            raise ValueError()

        return int(value[len('_integer['):-1])

    @staticmethod
    def _encode_str(value: str) -> str:
        """Encodes a (dotted) string
        Args:
            value: The value

        Returns:
            The encoded value
        """

        return value.replace('.', '\\u002e')

    @staticmethod
    def _decode_str(value: str) -> str:
        """Decodes a (dotted) string

        Args:
            value: The value

        Returns:
            The decoded value
        """

        return value.replace('\\u002e', '.')

    @staticmethod
    def _encode_data(data: Any) -> Any:
        """Recursively encode the data and apply dot replacement and integer encoding on the keys

        Args:
            data: The data
        Returns:
            The transformed data
        """

        if isinstance(data, dict):
            return {
                StorageMongoDb._encode_str(StorageMongoDb._encode_int(key) if isinstance(key, int) else key)
                : StorageMongoDb._encode_data(value)
                for key, value in data.items()
            }

        elif isinstance(data, list):
            return [StorageMongoDb._encode_data(item) for item in data]

        return data

    @staticmethod
    def _decode_data(data: Any) -> Any:
        """Recursively decode the data and apply dot replacement and integer decoding on the keys

        Args:
            data: The data
        Returns:
            The transformed data
        """

        if isinstance(data, dict):
            return {
                StorageMongoDb._decode_int(StorageMongoDb._decode_str(key))
                if StorageMongoDb._is_encoded_int(key) else StorageMongoDb._decode_str(key)
                : StorageMongoDb._decode_data(value)
                for key, value in data.items()
            }

        elif isinstance(data, list):
            return [StorageMongoDb._decode_data(item) for item in data]

        return data
